//OverSeer For .Net
//Modified by SharpPlus software (http://www.sharpplus.com)
//2004-11-14

{==============================================================================|
| Project : Delphree - Nexus - Overseer                          | 001.003.000 |
|==============================================================================|
| Content: Overseer client.                                                    |
|==============================================================================|
| The contents of this file are subject to the Mozilla Public License Ver. 1.0 |
| (the "License"); you may not use this file except in compliance with the     |
| License. You may obtain a copy of the License at http://www.mozilla.org/MPL/ |
|                                                                              |
| Software distributed under the License is distributed on an "AS IS" basis,   |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for |
| the specific language governing rights and limitations under the License.    |
|==============================================================================|
| The Original Code is Overseer for Nexus Delphi Library.                      |
|==============================================================================|
| The Initial Developer of the Original Code is Pavel Cisar (Czech Republic).  |
| Portions created by Pavel Cisar are Copyright (C) 1998, 1999, 2000.          |
| All Rights Reserved.                                                         |
|==============================================================================|
| Contributor(s):                                                              |
| Franz-Leo Chomse <chomse@compuserve.com>                                     |
|   PropertyToString, ObjectToStringList methods                               |
|==============================================================================|
| History: 1.2.1999 Version 001.000.000 (Pavel Cisar)                          |
|          First public release.                                               |
|          7.2.1999 Version 001.001.000 (Pavel Cisar)                          |
|          CodeSite(tm) segregation.                                           |
|          20.2.1999 Version 001.002.000 (Pavel Cisar)                         |
|          Make debugger channel ready for custom channels. Minor improvements.|
|          17.2.2000 Version 001.003.000 (Pavel Cisar)                         |
|          TNxDebugger Log routines changed to functions.                      |
|          Bug in SendString causing AV in case of empty string fixed.         |
|          LogProperty and LogObject are finally implemented.                  |
|          TNxDebuggerChannel class renamed to TNxOverseerChannel              |
|          TNxLogChannel class added                                           |
|          Overseer is now started in taskbar mode by default                  |
|==============================================================================}

unit uDbg;

interface

uses Windows, SysUtils, Messages, Forms, Registry, Dialogs,
  Classes, Graphics, uDbgIntf, Borland.Vcl.WinUtils, System.Runtime.InteropServices, System.Text;

//------------------------------------------------------------------------------
// Debugging Engine                                                            |
//------------------------------------------------------------------------------

type
  TNxDebugger = class;

  TNxCustomDebuggerChannel = class(TObject)
  private
    FDbgList: TList;
  protected
    function GetActive: boolean; virtual; abstract;
    procedure SetActive(Value: boolean); virtual;
    procedure RegisterDebugger(Dbg: TNxDebugger);
    procedure UnregisterDebugger(Dbg: TNxDebugger);
    procedure SendSections;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Send(ADbg: TNxDebugger; AMsg: TNxDebugMessageDesc); virtual; abstract;
    procedure StartDebugSession; virtual; abstract;
    procedure EndDebugSession; virtual; abstract;
    property Active: boolean read GetActive write SetActive;
  end;

  TNxLogMode = (lmOverride, lmIncrement);

  TNxLogChannel = class(TNxCustomDebuggerChannel)
  private
    FLogFile: TFileStream;
    FWriter: TWriter;
    FFileName: string;
    FMode: TNxLogMode;
  protected
    function GetActive: boolean; override;
    procedure SetActive(Value: boolean); override;
  public
    constructor Create(AMode: TNxLogMode; AFileName: string = '');
    destructor Destroy; override;
    procedure Send(ADbg: TNxDebugger; AMsg: TNxDebugMessageDesc); override;
    procedure StartDebugSession; override;
    procedure EndDebugSession; override;
  end;

  TNxOverseerChannel = class(TNxCustomDebuggerChannel)
  private
    FViewerWnd: hWnd;
    FDebuggerWnd: hWnd;
    FInterceptMessage: cardinal;
    function AllocateWindow: HWND;
  protected
    function GetActive: boolean; override;
    function WndProc(Window: HWND; Message, wParam, lParam: Longint): Longint;// stdcall;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Send(ADbg: TNxDebugger; AMsg: TNxDebugMessageDesc); override;
    procedure StartDebugSession; override;
    procedure EndDebugSession; override;
    property ViewerWindow: hWnd read FViewerWnd write FViewerWnd;
  end;

  TNxDebugger = class(TObject)
  private
    FID: string;
    FName: string;
    FAutoStart: boolean;
    FEnabled: boolean;
  protected
    procedure Send(AMsg: TNxDebugMessageDesc);
    function PointToString(APoint: TPoint): string;
    function RectToString(ARect: TRect): string;
    function CheckEnabled: boolean;
    function PropertyToString(PropOwner: TObject; PropName: string; var PropObject: TObject): string;
    procedure ObjectToStringList(Obj: TObject; AList: TStrings);
  public
    constructor Create(ID, Name: string);
    destructor Destroy; override;
    function LogMsg(AMsg: string): boolean;
    function LogMsgEx(AMsgType: integer; AMsg: string): boolean;
    function LogFmtMsg(Fmt: string; Args: array of const): boolean;
    function LogFmtMsgEx(AMsgType: integer; Fmt: string; Args: array of const): boolean;
    function LogNote(AMsg: string): boolean;
    function LogFmtNote(Fmt: string; Args: array of const): boolean;
    function LogError(AMsg: string): boolean;
    function LogFmtError(Fmt: string; Args: array of const): boolean;
    function LogWarning(AMsg: string): boolean;
    function LogFmtWarning(Fmt: string; Args: array of const): boolean;
    function LogAssigned(AMsg: string; Value: IntPtr{pointer}): boolean;
    function LogAssignedEx(AMsgType: integer; AMsg: string; Value: IntPtr{pointer}): boolean;
    function LogBoolean(AMsg: string; Expresion: boolean): boolean;
    function LogBooleanEx(AMsgType: integer; AMsg: string; Expresion: boolean): boolean;
    function LogColor(AMsg: string; Value: TColor): boolean;
    function LogColorEx(AMsgType: integer; AMsg: string; Value: TColor): boolean;
    function LogInteger(AMsg: string; Value: integer): boolean;
    function LogIntegerEx(AMsgType: integer; AMsg: string; Value: integer): boolean;
    function LogPoint(AMsg: string; Value: TPoint): boolean;
    function LogPointEx(AMsgType: integer; AMsg: string; Value: TPoint): boolean;
    function LogRect(AMsg: string; Value: TRect): boolean;
    function LogRectEx(AMsgType: integer; AMsg: string; Value: TRect): boolean;
    function LogString(AMsg: string; Value: string): boolean;
    function LogStringEx(AMsgType: integer; AMsg: string; Value: string): boolean;
    function LogProperty(AMsg: string; Obj: TObject; PropName: string): boolean;
    function LogObject(AMsg: string; Obj: TObject): boolean;
    function LogComponent(AMsg: string; Obj: TComponent): boolean;
    function LogStringList(AMsg: string; List: TStrings): boolean;
    function AddCheckPoint(CheckName: string): boolean;
    function AddSeparator: boolean;
    function Clear: boolean;
    function ScratchPad(LineIndex: integer; Msg: string): boolean;
    function FmtScratchPad(LineIndex: integer; Fmt: string; Args: array of const): boolean;
    function EnterProc(AMsg: string): boolean;
    function LeaveProc(AMsg: string): boolean;
    property Enabled: boolean read FEnabled;
  end;

procedure NxStartDebug;
procedure NxEndDebug;
function NxDebugging: boolean;

var
  NxDebuggerChannel: TNxCustomDebuggerChannel;

function Debugger:TNxDebugger;

implementation

uses ComObj, TypInfo;

var
  FDebugger: TNxDebugger;

resourcestring
  NxDelphreeRootKey = '\Software\Delphree';
  NxNexusKey = '\Nexus';
  NxDebugKey = '\Debug';

var
  DebuggerClosing: Boolean = False;
  DebuggerListenMessage: longint;
  nxNullGuid: TGUID = '{00000000-0000-0000-0000-000000000000}';//(D1: 0; D2: 0; D3: 0; D4: (0, 0, 0, 0, 0, 0, 0, 0));

//------------------------------------------------------------------------------
// Debugging Engine                                                            |
//------------------------------------------------------------------------------

procedure NxStartDebug;
begin
  NxDebuggerChannel.Active := true;
end;

procedure NxEndDebug;
begin
  NxDebuggerChannel.Active := false;
end;

function NxDebugging: boolean;
begin
  Result := NxDebuggerChannel.Active;
end;

//function DebuggerWndProc(Window: HWND; Message, wParam, lParam: Longint): LRESULT;//Longint;
function DebuggerWndProc(Window: HWND; Message: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT;
  //stdcall;
begin
  if Message = DebuggerListenMessage then
    if (NxDebuggerChannel<>nil) then
    begin
      Result := (NxDebuggerChannel as TNxOverseerChannel).WndProc(Window, Message,
        wParam, lParam);
      Exit;
    end;
  Result := DefWindowProc(Window, Message, wParam, lParam);
end;

var
  DebuggerWindowClass: TWndClass = (
    style: 0;
    lpfnWndProc: @DebuggerWndProc;
    cbClsExtra: 0;
    cbWndExtra: 0;
    hInstance: 0;
    hIcon: 0;
    hCursor: 0;
    hbrBackground: 0;
    lpszMenuName: '';//nil;
    lpszClassName: 'TDebuggerClientWindow');

{ TNxCustomDebuggerChannel }

constructor TNxCustomDebuggerChannel.Create;
begin
  inherited;
  FDbgList := TList.Create;
end;

destructor TNxCustomDebuggerChannel.Destroy;
var
  I: integer;
begin
  for I := 0 to FDbgList.Count - 1 do
    TNxDebugger(FDbgList.Items[I]).FEnabled := false;
  FDbgList.Free;
  if Active then
    EndDebugSession;
  inherited;
end;

procedure TNxCustomDebuggerChannel.RegisterDebugger(Dbg: TNxDebugger);
begin
  FDbgList.Add(Dbg);
  Dbg.FEnabled := Active;
end;

procedure TNxCustomDebuggerChannel.UnregisterDebugger(Dbg: TNxDebugger);
begin
  Dbg.FEnabled := false;
  FDbgList.Remove(Dbg);
end;

procedure TNxCustomDebuggerChannel.SendSections;
var
  M: TNxDebugMessageDesc;
  S: TStringList;
  I: integer;
begin
  S := TStringList.Create;
  try
    for I := 0 to FDbgList.Count - 1 do
      S.Values[TNxDebugger(FDbgList.Items[I]).FID] := TNxDebugger(FDbgList.Items[I]).FName;
    with M do
    begin
      MsgType := dbgSections;
      MsgText := S.Text;
      AttachmentType := NxNullGUID;
      AttachmentSize := 0;
    end;
  finally
    S.Free;
  end;
  Send(nil, M);
end;

procedure TNxCustomDebuggerChannel.SetActive(Value: boolean);
var
  R: TRegistry;
  S: string;
  H: HWND;
  I: integer;
begin
  if Value <> Active then
    if Value then
    begin
      StartDebugSession;
      if not Active then
      begin
       // Start debugger
        R := TRegistry.Create;
        try
          R.OpenKey(NxDelphreeRootKey + NxNexusKey + NxDebugKey, true);
          if R.ValueExists('Viewer') then
          begin
            S := R.ReadString('Viewer') + ' /taskbar';
            if S <> '' then
            begin
              I := 0; H := 0;
              if WinExec(S, SW_SHOW) > 31 then
                while (H = 0) and (I < 1000) do
                begin
                  Application.ProcessMessages;
                  H := FindWindow('TfrmDebugger', nil);
                  inc(I);
                end;
            end;
            StartDebugSession;
          end;
        finally
          R.Free;
        end;
      end;
    end
    else
      EndDebugSession;
end;

{ TNxOverseerChannel }

constructor TNxOverseerChannel.Create;
begin
  inherited;
  FDebuggerWnd := AllocateWindow;
  FInterceptMessage := RegisterWindowMessage('NxDbgViewer');
  DebuggerListenMessage := RegisterWindowMessage('NxDbgClient');
end;

destructor TNxOverseerChannel.Destroy;
begin
  DestroyWindow(FDebuggerWnd);
  inherited;
end;

function TNxOverseerChannel.AllocateWindow: HWND;
var
  //TempClass: TWndClass;
  TempClassInfo:TWndClassInfo;
  ClassRegistered: Boolean;
begin
  DebuggerWindowClass.hInstance := HInstance;
  ClassRegistered := GetClassInfo(HInstance, DebuggerWindowClass.lpszClassName,
    TempClassInfo);//TempClass);
  if not ClassRegistered {or (TempClassInfo.lpfnWndProc <> @DebuggerWndProc)} then
  begin
    if ClassRegistered then
      Windows.UnregisterClass(DebuggerWindowClass.lpszClassName, HInstance);
    Windows.RegisterClass(DebuggerWindowClass);
  end;
  Result := CreateWindow(DebuggerWindowClass.lpszClassName, '', 0,
    0, 0, 0, 0, 0, 0, HInstance, nil);
end;

type
  PBooleanArray = ^TBooleanArray;
  TBooleanArray = array[1..1000] of boolean;

function TNxOverseerChannel.WndProc(Window: HWND; Message, wParam, lParam: Longint): Longint;
begin
  Result := 0;
  if Message = DebuggerListenMessage then
    case wParam of
      dmsgStartSession: FViewerWnd := lParam;
      dmsgEndSession: FViewerWnd := 0;
      dmsgEnableSection: TNxDebugger(FDbgList.Items[lParam]).FEnabled := true;
      dmsgDisableSection: TNxDebugger(FDbgList.Items[lParam]).FEnabled := false;
    else
      Result := 1;
    end;
end;

function TNxOverseerChannel.GetActive: boolean;
begin
  Result := FViewerWnd <> 0;
end;

procedure TNxOverseerChannel.StartDebugSession;
var
  M: TNxDebugMessageDesc;
begin
  SendMessage(HWND_BROADCAST, FInterceptMessage, dmsgStartSession, FDebuggerWnd);
  with M do
  begin
    MsgType := dbgClientData;
    MsgText := Application.ExeName;
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(nil, M);
  SendSections;
end;

procedure TNxOverseerChannel.EndDebugSession;
var
  I: integer;
begin
  SendMessage(FViewerWnd, FInterceptMessage, dmsgEndSession, FDebuggerWnd);
  for I := 0 to FDbgList.Count - 1 do
    TNxDebugger(FDbgList.Items[I]).FEnabled := false;
end;

procedure TNxOverseerChannel.Send(ADbg: TNxDebugger; AMsg: TNxDebugMessageDesc);
var
  M: IntPtr;//PNxDebugMessage;
  DebugMsg:TNxDebugMessage;
  I: integer;
  CDS: TCopyDataStruct;
  PParam:IntPtr;
  index:integer;
  s:string;
  b:array of byte;
begin
  if not Active then exit;
  CDS.cbData := sizeof(TNxDebugMessage) + length(AMsg.MsgText) + 1 + AMsg.AttachmentSize;
  M:=Marshal.AllocHGlobal(CDS.cbData);
  DebugMsg.Section:=FDbgList.IndexOf(ADbg);
  DebugMsg.MsgType:=AMsg.MsgType;
  DebugMsg.MsgTime:=Now;//38005.891881;
  DebugMsg.MsgSize := Encoding.GetEncoding(0).GetByteCount(AMsg.MsgText);//length(AMsg.MsgText);
  DebugMsg.AttachmentType:=AMsg.AttachmentType;
  DebugMsg.AttachmentSize:=AMsg.AttachmentSize;
  DebugMsg.Data:=Convert.ToByte('z');

  Marshal.StructureToPtr(TObject(DebugMsg), M,false);
  //Write string to hglobal
  index:=sizeof(TNxDebugMessage)-4;//Todo:why we need to substract 4?
  b:=Encoding.GetEncoding(0).GetBytes(AMsg.MsgText);
  for I:=0 to DebugMsg.MsgSize-1 do
  begin
    Marshal.WriteByte(M,index,b[i]);
    inc(index);
  end;
  //Done: implement attachment
  if AMsg.AttachmentSize>0 then
  begin
    s:=Marshal.PtrToStringAnsi(AMsg.Attachment);
    b:=Encoding.GetEncoding(0).GetBytes(s);
    for I:=0 to AMsg.AttachmentSize-2 do
    begin
      Marshal.WriteByte(M,index, b[i]);
      inc(index);
    end;
    Marshal.WriteByte(M,index,0);
  end;
  PParam := Marshal.AllocHGlobal(SizeOf(CDS));
  try
    try
        CDS.lpData := M;
        Marshal.StructureToPtr(TObject(CDS), PParam, False);
        SendMessage(FViewerWnd, WM_COPYDATA, 0, LongInt(PParam));
    except on E:Exception do
      raise E;
    end;
  finally
    Marshal.FreeHGlobal(PParam);
    Marshal.FreeHGlobal(M);
  end;
end;

{ TNxDebugger }

constructor TNxDebugger.Create(ID, Name: string);
begin
  inherited Create;
  FID := ID;
  FName := Name;
  FAutoStart := true;
  NxDebuggerChannel.RegisterDebugger(self);
end;

destructor TNxDebugger.Destroy;
begin
  if (Self = Debugger) and (not DebuggerClosing) then
    LogError('Debugger destroyed too early.');
  NxDebuggerChannel.UnregisterDebugger(self);
  inherited;
end;

procedure TNxDebugger.Send(AMsg: TNxDebugMessageDesc);
begin
  if CheckEnabled then
    NxDebuggerChannel.Send(self, AMsg);
end;

type
  ColorTableRec = record
    Name: string;
    Color: TColor;
  end;

const
  MaxColor = 45;

function TNxDebugger.PointToString(APoint: TPoint): string;
begin
  Result := '(' + IntToStr(APoint.x) + ',' + IntToStr(APoint.y) + ')';
end;

function TNxDebugger.RectToString(ARect: TRect): string;
begin
  Result := '(' + PointToString(ARect.TopLeft) + ',' +
    PointToString(ARect.BottomRight) + ')';
end;

function TNxDebugger.CheckEnabled: boolean;
begin
  if (not NxDebuggerChannel.Active) and FAutoStart then
    NxDebuggerChannel.StartDebugSession;
  FAutoStart := false;
  Result := Enabled;
end;

function TNxDebugger.LogMsg(AMsg: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgInfo;
    MsgText := AMsg;
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogMsgEx(AMsgType: integer; AMsg: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := AMsgType;
    MsgText := AMsg;
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogFmtMsg(Fmt: string; Args: array of const): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgInfo;
    MsgText := Format(Fmt, Args);
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogFmtMsgEx(AMsgType: integer; Fmt: string; Args: array of const): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := AMsgType;
    MsgText := Format(Fmt, Args);
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogNote(AMsg: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgNote;
    MsgText := AMsg;
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogFmtNote(Fmt: string; Args: array of const): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgNote;
    MsgText := Format(Fmt, Args);
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogError(AMsg: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgError;
    MsgText := AMsg;
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogFmtError(Fmt: string; Args: array of const): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgError;
    MsgText := Format(Fmt, Args);
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogWarning(AMsg: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgWarning;
    MsgText := AMsg;
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogFmtWarning(Fmt: string; Args: array of const): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgWarning;
    MsgText := Format(Fmt, Args);
    AttachmentType := NxNullGUID;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogAssigned(AMsg: string; Value: IntPtr{pointer}): boolean;
var
  M: TNxDebugMessageDesc;
  S: string;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  if assigned(Value) then
    S := ' (Assigned)'
  else
    S := ' (Unassigned)';
  with M do
  begin
    MsgType := dbgInfo;
    MsgText := AMsg + S;
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogAssignedEx(AMsgType: integer; AMsg: string; Value: IntPtr{pointer}): boolean;
var
  M: TNxDebugMessageDesc;
  S: string;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  if assigned(Value) then
    S := ' (Assigned)'
  else
    S := ' (Unassigned)';
  with M do
  begin
    MsgType := AMsgType;
    MsgText := AMsg + S;
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogBoolean(AMsg: string; Expresion: boolean): boolean;
var
  M: TNxDebugMessageDesc;
  S: string;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  if Expresion then
    S := ' (TRUE)'
  else
    S := ' (FALSE)';
  with M do
  begin
    MsgType := dbgInfo;
    MsgText := AMsg + S;
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogBooleanEx(AMsgType: integer; AMsg: string; Expresion: boolean): boolean;
var
  M: TNxDebugMessageDesc;
  S: string;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  if Expresion then
    S := ' (TRUE)'
  else
    S := ' (FALSE)';
  with M do
  begin
    MsgType := AMsgType;
    MsgText := AMsg + S;
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogColor(AMsg: string; Value: TColor): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgInfo;
    MsgText := AMsg + ' ' + ColorToString(Value);
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogColorEx(AMsgType: integer; AMsg: string; Value: TColor): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := AMsgType;
    MsgText := AMsg;
    MsgText := AMsg + ' ' + ColorToString(Value);
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogInteger(AMsg: string; Value: integer): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgInfo;
    MsgText := AMsg + ' (' + IntToStr(Value) + ')';
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogIntegerEx(AMsgType: integer; AMsg: string; Value: integer): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := AMsgType;
    MsgText := AMsg + ' (' + IntToStr(Value) + ')';
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogPoint(AMsg: string; Value: TPoint): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgInfo;
    MsgText := AMsg + ' ' + PointToString(Value);
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogPointEx(AMsgType: integer; AMsg: string; Value: TPoint): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := AMsgType;
    MsgText := AMsg + ' ' + PointToString(Value);
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogRect(AMsg: string; Value: TRect): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgInfo;
    MsgText := AMsg + ' ' + RectToString(Value);
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogRectEx(AMsgType: integer; AMsg: string; Value: TRect): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := AMsgType;
    MsgText := AMsg + ' ' + RectToString(Value);
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LogString(AMsg: string; Value: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgInfo;
    MsgText := AMsg;
    AttachmentType := attPChar;
    AttachmentSize := System.Text.Encoding.GetEncoding(0).GetByteCount(Value);//length(Value);
    if AttachmentSize > 0 then
      inc(AttachmentSize);
    //Attachment :=Marshal.StringToHGlobalUni(value);//IntPtr(value);//pointer(Value);
    Attachment :=Marshal.StringToHGlobalAnsi(value);//IntPtr(value);//pointer(Value);
  end;
  Send(M);
end;

function TNxDebugger.LogStringEx(AMsgType: integer; AMsg: string; Value: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := AMsgType;
    MsgText := AMsg;
    AttachmentType := attPChar;
    AttachmentSize := length(Value) + 1;
    Attachment :=     Marshal.StringToHGlobalAnsi(value);//IntPtr(Value);//Pointer(Value);
  end;
  Send(M);
end;

function TNxDebugger.LogProperty(AMsg: string; Obj: TObject; PropName: string): boolean;
var
  M: TNxDebugMessageDesc;
  S: string;
  PropertyObject: TObject;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  S := PropertyToString(Obj, PropName, PropertyObject);
  with M do
  begin
    MsgType := dbgProperty;
    MsgText := AMsg;
    AttachmentType := attObjectProperties;
    AttachmentSize := length(S);
    if AttachmentSize > 0 then
      inc(AttachmentSize);
    Attachment :=     Marshal.StringToHGlobalAnsi(s);//IntPtr(S);//pointer(S);
  end;
  Send(M);
end;

function TNxDebugger.LogObject(AMsg: string; Obj: TObject): boolean;
var
  M: TNxDebugMessageDesc;
  S: TMemoryStream;
  List: TStringList;
begin
  Result := true;
(*  if not CheckEnabled then
    exit;
  List := TStringList.Create;
  try
    ObjectToStringList(Obj, List);
    S := TMemoryStream.Create;
    try
      List.SaveToStream(S);
      with M do
      begin
        MsgType := dbgObject;
        MsgText := AMsg;
        AttachmentType := attObjectProperties;
        AttachmentSize := S.Size;
        Attachment := S.Memory;
      end;
      Send(M);
    finally
      S.Free;
    end;
  finally
    List.Free;
  end;*)
end;

function TNxDebugger.LogComponent(AMsg: string; Obj: TComponent): boolean;
var
  M: TNxDebugMessageDesc;
  S: TMemoryStream;
begin
  Result := true;
(*  if not CheckEnabled then
    exit;
  S := TMemoryStream.Create;
  try
    S.WriteComponent(Obj);
    with M do
    begin
      MsgType := dbgStream;
      MsgText := AMsg;
      AttachmentType := attComponentStream;
      AttachmentSize := S.Size;
      Attachment := S.Memory;
    end;
    Send(M);
  finally
    S.Free;
  end;*)
end;

function TNxDebugger.LogStringList(AMsg: string; List: TStrings): boolean;
var
  M: TNxDebugMessageDesc;
  S: TMemoryStream;
begin
  Result := true;
{  if not CheckEnabled then
    exit;
  S := TMemoryStream.Create;
  try
    List.SaveToStream(S);
    with M do
    begin
      MsgType := dbgStringList;
      MsgText := AMsg;
      AttachmentType := attStringList;
      AttachmentSize := S.Size;
      Attachment := S.Memory;
    end;
    Send(M);
  finally
    S.Free;
  end;}
end;

function TNxDebugger.AddCheckPoint(CheckName: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgCheckPoint;
    MsgText := CheckName;
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.AddSeparator: boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgSeparator;
    MsgText := '-';
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.Clear: boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgClear;
    MsgText := '';
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.ScratchPad(LineIndex: integer; Msg: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
(*  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgBlackboard;
    MsgText := Msg;
    AttachmentType := NxNullGuid;
    AttachmentSize := sizeof(integer);
    Attachment := @LineIndex;
  end;
  Send(M);*)
end;

function TNxDebugger.FmtScratchPad(LineIndex: integer; Fmt: string; Args: array of const): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
(*  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgBlackboard;
    MsgText := Format(Fmt, Args);
    AttachmentType := NxNullGuid;
    AttachmentSize := sizeof(integer);
    Attachment := @LineIndex;
  end;
  Send(M);*)
end;

function TNxDebugger.EnterProc(AMsg: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgEnterMethod;
    MsgText := AMsg;
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.LeaveProc(AMsg: string): boolean;
var
  M: TNxDebugMessageDesc;
begin
  Result := true;
  if not CheckEnabled then
    exit;
  with M do
  begin
    MsgType := dbgExitMethod;
    MsgText := AMsg;
    AttachmentType := NxNullGuid;
    AttachmentSize := 0;
  end;
  Send(M);
end;

function TNxDebugger.PropertyToString(PropOwner: TObject; PropName: string;
  var PropObject: TObject): string;
type
  TIntegerSet = set of 0..SizeOf(Integer) * 8 - 1; // see CLASSES.PAS
var
  PropertyInfo: IntPtr;//PPropInfo;
  PropertyType: IntPtr;//PTypeInfo;
  PropertyKind: TTypeKind;
  BaseType: IntPtr;//PTypeInfo;
  BaseData: IntPtr;//PTypeData;
  GetProc: IntPtr;//Pointer;
  OrdValue: Integer;
  FloatValue: Extended;
  EnumName: string;
  N: Integer;
  R: Integer;
begin
  result:='Todo';
(*  PropObject := nil;
  PropertyInfo := GetPropInfo(PropOwner.ClassInfo, PropName);
  if Assigned(PropertyInfo) then
  begin
    PropertyType := PropertyInfo^.PropType^;
    GetProc := PropertyInfo^.GetProc;
    PropertyKind := PropertyType^.Kind;
    Result := '<' + PropertyType^.Name + '>';
    if not Assigned(GetProc) then
      Result := Result + ' (can''t be read)'
    else
    begin
      case PropertyKind of
        tkSet:
          begin
            BaseType := GetTypeData(PropertyType)^.CompType^;
            BaseData := GetTypeData(BaseType);
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := '[';
            R := 0;
            for N := BaseData^.MinValue to BaseData^.MaxValue do
              if N in TIntegerSet(OrdValue) then
              begin
                EnumName := GetEnumName(BaseType, N);
                if EnumName = '' then
                  Break;
                if R > 0 then
                  Result := Result + ', ';
                Result := Result + EnumName;
                inc(R);
              end;
            Result := Result + ']';
          end;
        tkInteger:
          begin
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := IntToStr(OrdValue);
          end;
        tkChar:
          begin
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := '#$' + IntToHex(OrdValue, 2);
          end;
        tkWChar:
          begin
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := '#$' + IntToHex(OrdValue, 4);
          end;
        tkClass:
          begin
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := '<' + PropertyType^.Name + '>';
            if OrdValue = 0 then
              Result := Result + ' (not assigned)'
            else
              Result := Result + ' (assigned)';
            PropObject := TObject(OrdValue);
          end;
        tkFloat:
          begin
            FloatValue := GetFloatProp(PropOwner, PropertyInfo);
            Result := FormatFloat('n', FloatValue);
          end;
        tkEnumeration:
          begin
            OrdValue := GetOrdProp(PropOwner, PropertyInfo);
            Result := GetEnumName(PropertyType, OrdValue);
          end;
        tkLString:
          Result := '''' + GetStrProp(PropOwner, PropertyInfo) + '''';
        tkWString:
          Result := '''' + GetStrProp(PropOwner, PropertyInfo) + '''';
        tkString:
          Result := '''' + GetStrProp(PropOwner, PropertyInfo) + '''';
        tkVariant:
          Result := GetVariantProp(PropOwner, PropertyInfo);
        tkMethod:
          Result := Result + '(' + GetEnumName(TypeInfo(TMethodKind),
            Ord(GetTypeData(PropertyType)^.MethodKind)) + ')';
      else
        Result := '<' + PropertyType^.Name + '> ('
          + GetEnumName(TypeInfo(TTypeKind), Ord(PropertyKind)) + ')';
      end
    end
  end
  else
    Result := '<unknown>';
  Result := PropName + ' = ' + Result;*)
end;

procedure TNxDebugger.ObjectToStringList(Obj: TObject; AList: TStrings);

(*  procedure AddObjectToStringList(PropOwner: TObject; List: TStrings; Level: Integer);
  type
    TIntegerSet = set of 0..SizeOf(Integer) * 8 - 1; // see CLASSES.PAS
  var
    PropIdx: integer;
    PropertyList: ^TPropList;
    PropertyName: string;
    PropertyInfo: PPropInfo;
    PropertyType: PTypeInfo;
    PropertyKind: TTypeKind;
    BaseType: PTypeInfo;
    BaseData: PTypeData;
    GetProc: Pointer;
    OrdValue: Integer;
    FloatValue: Extended;
    N: Integer;
    Prefix: string;
    NewLine: string;
    EnumName: string;
    NextObject: TObject;
    FollowObject: Boolean;
  begin
    GetMem(PropertyList, SizeOf(TPropList));
    try
      Prefix := StringOfChar(' ', 2 * Level);
      // Build list of published properties
      FillChar(PropertyList^[0], SizeOf(TPropList), #00);
      GetPropList(PropOwner.ClassInfo, tkProperties - [tkArray, tkRecord,
        tkInterface], @PropertyList^[0]);
      // Process property list
      PropIdx := 0;
      while ((PropIdx < High(PropertyList^)) and (nil <> PropertyList^[PropIdx])) do
      begin
        // Get information about found properties
        PropertyInfo := PropertyList^[PropIdx];
        PropertyType := PropertyInfo^.PropType^;
        PropertyKind := PropertyType^.Kind;
        PropertyName := PropertyInfo^.Name;
        // Write only property
        GetProc := PropertyInfo^.GetProc;
        if not Assigned(GetProc) then
        begin
          NewLine := Prefix + '  ' + PropertyName + ' = <' + PropertyType^.Name +
            '> (can''t be read)';
          List.Add(NewLine);
        end
        else
        begin
          case PropertyKind of
            tkSet:
              begin
                BaseType := GetTypeData(PropertyType)^.CompType^;
                BaseData := GetTypeData(BaseType);
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '+ ' + PropertyName + ' = [' + BaseType^.Name
                  + ']';
                List.Add(NewLine);
                for N := BaseData^.MinValue to BaseData^.MaxValue do
                begin
                  EnumName := GetEnumName(BaseType, N);
                  if EnumName = '' then
                    Break;
                  NewLine := Prefix + '    ' + EnumName;
                  if N in TIntegerSet(OrdValue) then
                    NewLine := NewLine + ' = True'
                  else
                    NewLine := NewLine + ' = False';
                  List.Add(NewLine);
                end;
              end;
            tkInteger:
              begin
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '  ' + PropertyName + ' = ' + IntToStr(OrdValue);
                List.Add(NewLine);
              end;
            tkChar:
              begin
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '  ' + PropertyName + ' = ' + '#$' +
                  IntToHex(OrdValue, 2);
                List.Add(NewLine);
              end;
            tkWChar:
              begin
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '  ' + PropertyName + ' = #$' + IntToHex(OrdValue, 4);
                List.Add(NewLine);
              end;
            tkClass:
              begin
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                if OrdValue = 0 then
                begin
                  NewLine := Prefix + '  ' + PropertyName + ' = <' +
                    PropertyType^.Name + '> (not assigned)';
                  List.Add(NewLine);
                end
                else
                begin
                  NextObject := TObject(OrdValue);
                  NewLine := Prefix + '  ' + PropertyName + ' = <' +
                    PropertyType^.Name + '>';
                  if NextObject is TComponent then
                  begin
                    FollowObject := False;
                    NewLine := NewLine + ': ' + TComponent(NextObject).Name
                  end
                  else
                  begin
                    FollowObject := True;
                    NewLine[Succ(Length(Prefix))] := '*';
                  end;
                  List.Add(NewLine);
                  if FollowObject then
                  begin
                    try
                      AddObjectToStringList(NextObject, List, Level + 1);
                    except
                      List.Add('*** Exception triggered ***');
                    end;
                  end;
                end;
              end;
            tkFloat:
              begin
                FloatValue := GetFloatProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '  ' + PropertyName + ' = ' +
                  FormatFloat('n', FloatValue);
                List.Add(NewLine);
              end;
            tkEnumeration:
              begin
                OrdValue := GetOrdProp(PropOwner, PropertyInfo);
                NewLine := Prefix + '  ' + PropertyName + ' = ' +
                  GetEnumName(PropertyType, OrdValue);
                List.Add(NewLine);
              end;
            tkLString:
              begin
                NewLine := Prefix + '  ' + PropertyName + ' = ' + '''' +
                  GetStrProp(PropOwner, PropertyInfo) + '''';
                List.Add(NewLine);
              end;
            tkWString:
              begin
                NewLine := Prefix + '  ' + PropertyName + ' = ' + '''' +
                  GetStrProp(PropOwner, PropertyInfo) + '''';
                List.Add(NewLine);
              end;
            tkString:
              begin
                NewLine := Prefix + '  ' + PropertyName + ' = ' + '''' +
                  GetStrProp(PropOwner, PropertyInfo) + '''';
                List.Add(NewLine);
              end;
            tkVariant:
              begin
                NewLine := Prefix + '  ' + PropertyName + ' = ' +
                  GetVariantProp(PropOwner, PropertyInfo);
                List.Add(NewLine);
              end;
            tkMethod:
              begin
                NewLine := Prefix + '  ' + PropertyName + ' = (' +
                  GetEnumName(TypeInfo(TMethodKind),
                  Ord(GetTypeData(PropertyType)^.MethodKind)) + ')';
                List.Add(NewLine);
              end;
          else
            begin
              NewLine := Prefix + '  ' + PropertyName + ' = <' +
                PropertyType^.Name + '> ('
                + GetEnumName(TypeInfo(TTypeKind), Ord(PropertyKind)) + ')';
              List.Add(NewLine);
            end;
          end
        end;
        // Next item in the property list
        Inc(PropIdx);
      end;
      NewLine := '';
    finally
      if NewLine <> '' then
        List.Add(NewLine);
      FreeMem(PropertyList);
    end;
  end;*)

begin
(*  try
    AddObjectToStringList(Obj, AList, 0)
  except
    AList.Add('*** Exception triggered ***');
  end;*)
end;

{ TNxLogChannel }

constructor TNxLogChannel.Create(AMode: TNxLogMode; AFileName: string = '');
begin
  inherited Create;
  if AFileName = '' then
    FFileName := ChangeFileExt(Application.ExeName, '.LOG')
  else
    FFileName := AFileName;
  FMode := AMode;
end;

destructor TNxLogChannel.Destroy;
begin
  EndDebugSession;
  inherited;
end;

procedure TNxLogChannel.EndDebugSession;
begin
  if Active then
  begin
    FWriter.WriteListEnd;
    FWriter.FlushBuffer;
    FreeAndNil(FWriter);
    FreeAndNil(FLogFile);
  end;
end;

function TNxLogChannel.GetActive: boolean;
begin
  Result := assigned(FLogFile);
end;

procedure TNxLogChannel.Send(ADbg: TNxDebugger; AMsg: TNxDebugMessageDesc);
var
  T: TDateTime;
begin
  FWriter.WriteInteger(FDbgList.IndexOf(ADbg));
  FWriter.WriteInteger(AMsg.MsgType);
  FWriter.WriteInteger(0);
  FWriter.WriteString(AMsg.MsgText);
  T := Now;
  FWriter.Write(T, sizeof(TDateTime));
  //FWriter.Write(AMsg.AttachmentType, sizeof(TGUID));
  FWriter.write(AMsg.AttachmentType.ToByteArray(),sizeof(TGuid));
  FWriter.WriteInteger(AMsg.AttachmentSize);
 // if AMsg.AttachmentSize > 0 then
 //   FWriter.write(AMsg.Attachment, AMsg.AttachmentSize);
end;

procedure TNxLogChannel.SetActive(Value: boolean);
begin
  if Value <> Active then
  begin
    if Value then
      StartDebugSession
    else
      EndDebugSession;
  end;
end;

procedure TNxLogChannel.StartDebugSession;
var
  M: word;
  S: string;
  FName: string;
  FPath: string;
  FExt: string;
  I: integer;
begin
  EndDebugSession;
  M := fmCreate;
  case FMode of
    lmOverride:
      begin
        M := fmCreate;
        S := FFileName;
      end;
    lmIncrement:
      begin
        FPath := ExtractFilePath(FFileName);
        FExt := ExtractFileExt(FFileName);
        FName := ExtractFileName(FFileName);
        FName := copy (FName,1,length(FName)-length(FExt));
        S := FPath+FName+FExt;
        M := 1;
        while FileExists(S) do
        begin
          S := FPath+FName+IntToStr(M)+FExt;
          inc(M);
        end;
        M := fmCreate;
      end;
  end;
  FLogFile := TFileStream.Create(S, M);
  FWriter := TWriter.Create(FLogFile, 1024);
  FWriter.WriteString(NxLogSign);
  FWriter.WriteListBegin;
  for I := 0 to FDbgList.Count - 1 do
    TNxDebugger(FDbgList.Items[I]).FEnabled := true;
end;

function Debugger:TNxDebugger;
begin
  if (FDebugger=nil) then
    FDebugger := TNxDebugger.Create('General', 'All nonspecific messages');
  result:=FDebugger;
end;

initialization
  NxDebuggerChannel := TNxOverseerChannel.Create;
// To test Log facility via DbgDemo demo application, comment out line above and
// uncomment next line. Log will be re-routed to DbgDemo.LOG file.
//  NxDebuggerChannel := TNxLogChannel.Create (lmOverride);


finalization
  //DebuggerClosing := True;
  //Debugger.Free;
  //NxDebuggerChannel.Free;

end.

